/**************************** CPP SOURCE FILE ********************************
**
**  Project:    RTCU Communication Hub, Namelist plug-in
**  Filename:   NAMELIST.CPP 
**  Version:    1.00
**  Date:       January 30, 2019
**  
******************************************************************************
**
**  VERSION HISTORY:
**  ----------------
**  1.00 - January 30, 2019 - Michael Olsen
**  - Ported to RCH.
**
*****************************************************************************/


/*****************************************************************************
**
** MODULES USED 
**
*****************************************************************************/
#include <Windows.h>
#include <process.h>
#include <map>
#include <string>
#include "local.h"
#include "tinyxml2.h"


/*****************************************************************************
**
**    DEFINITIONS AND MACROS
**
*****************************************************************************/


/****************************************************************************
**
**    TYPEDEFS AND STRUCTURES
**
****************************************************************************/
typedef std::map<long, std::string>     node_map_t;
typedef node_map_t::iterator            node_iter_t;


/*****************************************************************************
**
** EXPORTED VARIABLES 
**
*****************************************************************************/


/*****************************************************************************
**
** GLOBAL VARIABLES 
**
*****************************************************************************/
static node_map_t                       list_;
static CRITICAL_SECTION                 lock_;
static std::string                      filename_;
static long                             delay_ = 6000; // Default 10 min.
static node_iter_t                      read_ = list_.end();
static bool                             iterating_ = false;
static bool                             running_ = false;
static bool                             halt_ = false;


/*****************************************************************************
**
** LOCAL FUNCTIONS 
**
*****************************************************************************/

/////////////////////////////////////////////////////////////////////////////
// This function will convert an UTF-8 string to an ANSI string.
//
// Input:
//    text          - The UTF-8 string.
//
// Output:
//    None.
//
// Returns:
//    The ANSI string.
//
/////////////////////////////////////////////////////////////////////////////
static std::string Utf8ToAnsi(const std::string &text)
{
  // Var
  std::string  res;
  wchar_t*     temp;
  char*        data;
  int          size, len;

  // Determine the size of the resulting string
  size = MultiByteToWideChar(CP_UTF8, 0, text.c_str(), text.length(), nullptr, 0);
  if (size < 1) return res;

  // Allocate temporary buffer
  temp = new wchar_t[size + 1];
  if (temp == nullptr) return res;

  // Convert text
  MultiByteToWideChar(CP_UTF8, 0, text.c_str(), text.length(), temp, size);

  // Determine the size of the resulting string
  len = WideCharToMultiByte(CP_ACP, 0, temp, size, nullptr, 0, nullptr, nullptr);

  // Allocate temporary buffer
  data = new char[len + 1];
  if (data == nullptr) {
    delete [] temp;
    return res;
  }

  // Convert text
  WideCharToMultiByte(CP_ACP, 0, temp, size, data, len + 1, nullptr, nullptr);
  res = std::string(data, len);

  // Cleanup
  delete [] temp;
  delete [] data;

  // Completed
  return res;
}

/////////////////////////////////////////////////////////////////////////////
// This function will read a value from a XML node.
//
// Input:
//    None.
//
// Output:
//    None.
//
// Returns:
//    The value read or zero.
//
/////////////////////////////////////////////////////////////////////////////
static long ReadNodeValue(const tinyxml2::XMLElement *pItem)
{
  // Validate
  if (pItem == nullptr) return 0;
  const char * text = pItem->GetText();
  if (text == nullptr) return 0;

  // Convert the value
  return atol(text);
}

/////////////////////////////////////////////////////////////////////////////
// This function will read the text from a XML node.
//
// Input:
//    None.
//
// Output:
//    None.
//
// Returns:
//    A string that holds the node text or is empty.
//
/////////////////////////////////////////////////////////////////////////////
static std::string ReadNodeText(const tinyxml2::XMLElement *pItem)
{
  // Validate
  if (pItem == nullptr) return std::string();
  const char * text = pItem->GetText();
  if (text == nullptr) return std::string();

  // Convert the value
  return std::string(text);
}

/////////////////////////////////////////////////////////////////////////////
// This function will update the list of denied Node ID's.
//
// Input:
//    None.
//
// Output:
//    None.
//
// Returns:
//    0             - List is updated.
//    1             - File is not found.
//    2             - File parsing error.
//    3             - File is not a blacklist data file.
//
/////////////////////////////////////////////////////////////////////////////
int UpdateList(void)
{
  // Var
  tinyxml2::XMLDocument config_;
  tinyxml2::XMLElement  *pItem, *pConfig;
  node_map_t            temp;
  node_iter_t           it_cur, it_new;
  int                   rc;

  // Load blacklisted nodes
  rc = config_.LoadFile(filename_.c_str());
  if (rc == tinyxml2::XML_ERROR_FILE_NOT_FOUND ||
      rc == tinyxml2::XML_ERROR_FILE_COULD_NOT_BE_OPENED) return 1;
  if (rc != 0) return 2;
  pConfig = config_.FirstChildElement("namelist");
  if (pConfig == nullptr) return 3;

  // Iterate nodes
  pItem = pConfig->FirstChildElement("node");
  while (pItem) {
    // Insert
    temp.insert(node_map_t::value_type(ReadNodeValue(pItem->FirstChildElement("id")),
                                       ReadNodeText(pItem->FirstChildElement("name"))));

    // Next node
    pItem = pItem->NextSiblingElement();
  }

  // Lock list
  EnterCriticalSection(&lock_);

  // Iterate list
  for (it_cur = list_.begin(), it_new = temp.begin();
       it_cur != list_.end() && it_new != temp.end();
       it_cur++, it_new++) {
    if (it_cur->first != it_new->first || it_cur->second.compare(it_new->second) != 0) {
      break;
    }
  }
  if (it_cur != list_.end() || it_new != temp.end()) {
    // The list is updated
    write_log(3, "Namelist is refreshed.");
  }

  // Update list
  list_.clear();
  list_.insert(temp.begin(), temp.end());
  read_ = list_.end();

  // Unlock list
  LeaveCriticalSection(&lock_);

  // Completed
  return 0;
}

/////////////////////////////////////////////////////////////////////////////
// This thread handles the timer.
//
// Input:
//      None.
//
// Output:
//      None.
//
// Returns:
//      None.
//
/////////////////////////////////////////////////////////////////////////////
static void update_thread(void* /*arg*/)
{
  // Var
  long interval = delay_;
  int  rc;

  // Mark as running
  running_ = true;

  // Do forever
  for (;;) {
    // Wait
    while (!halt_ && --interval > 0) Sleep(100);
    if (halt_) break;
    interval = delay_;

    // Update
    rc = UpdateList();
    if (rc == 1) write_log(1, "Namelist XML file is missing!");
    else if (rc) write_log(1, "Namelist XML file parsing error!");
  }

  // Mark as stopped
  running_ = false;
}


/*****************************************************************************
**
** EXPORTED FUNCTIONS 
**
*****************************************************************************/

/////////////////////////////////////////////////////////////////////////////
// This function will initialize the name list functionality.
//
// Input:
//    config        - Pointer to the configuration string from the RCH.
//
// Output:
//    None.
//
// Returns:
//    0             - Plug-in is initialized.
//    1             - Illegal configuration string.
//    2             - No file specified in configuration.
//    3             - XML file with namelist is not found.
//    4             - Error parsing XML file.
//
/////////////////////////////////////////////////////////////////////////////
int InitNamelist(const char *config)
{
  // Var
  tinyxml2::XMLDocument       config_;
  tinyxml2::XMLElement        *pItem, *pConfig;
  int                         rc;

  // Initialize thread
  InitializeCriticalSection(&lock_);

  // Load configuration
  if (config_.Parse(config) != 0) return 1;
  pConfig = config_.FirstChildElement("configuration");
  if (pConfig == nullptr) return 1;

  // Iterate properties
  pItem = pConfig->FirstChildElement("property");
  while (pItem) {
    // Filename
    if (_stricmp(pItem->Attribute("name"), "File") == 0) {
      filename_ = Utf8ToAnsi(ReadNodeText(pItem));
    }
    // Time interval
    else if (_stricmp(pItem->Attribute("name"), "Refresh interval") == 0) {
      delay_ = ReadNodeValue(pItem);
      if (delay_ == 0) delay_ = 10;
      delay_ *= 600; // Convert from minutes to number of 100 ms ticks.
    }

    // Next property
    pItem = pItem->NextSiblingElement();
  }
  if (filename_.empty()) return 2;

  // Start thread
  halt_ = false;
  _beginthread(update_thread, 0, nullptr);

  // Load namelisted nodes
  rc = UpdateList();
  if (rc == 1) {
     write_log(1, "Namelist XML file is missing!");
     return 3;
  }
  if (rc) {
     write_log(1, "Namelist XML file parsing error!");
     return 4;
  }

  // Completed
  return 0;
}

/////////////////////////////////////////////////////////////////////////////
// This function will stop the blacklist and cleanup resources.
//
// Input:
//    None.
//
// Output:
//    None.
//
// Returns:
//    None.
//
/////////////////////////////////////////////////////////////////////////////
void StopNamelist(void)
{
  // Var
  int count = 50;

  // Stop thread
  halt_ = true;
  while (running_ && count-- > 0) Sleep(100);

  // Cleanup
  list_.clear();
  
  // Cleanup thread
  DeleteCriticalSection(&lock_);
}

/////////////////////////////////////////////////////////////////////////////
// This function is called from the RCH when starting to iterate the
// names in the list. The list will be locked until iteration is completed.
//
// Input:
//    name_len      - Pointer to a variable that holds the length of the
//                    name buffer in Bytes. If the buffer is to small it
//                    will hold the size required.
//
// Output:
//    name          - Pointer to the buffer where the name is stored.
//                    The name is a zero terminated UTF-8 string.
//    node          - Pointer to the buffer where the Node ID is stored.
//
// Returns:
//    0             - The name and node id is returned.
//    1             - One or more of the parameters are missing.
//    2             - The name buffer is not large enough.
//    3             - No more names in list.
//
/////////////////////////////////////////////////////////////////////////////
int GetFirstName(char *name, int *name_len, long *node)
{
  // Var
  long          val_node;
  std::string   val_name;

  // Verify inputs
  if (name == nullptr) return 1;
  if (name_len == nullptr) return 1;
  if (node == nullptr) return 1;

  // Lock list
  EnterCriticalSection(&lock_);

  // Verify list
  if (list_.empty()) {
    // Unlock list
    LeaveCriticalSection(&lock_);
    // No names in list
    return 3;
  }

  // Get first entry
  read_      = list_.begin();
  val_node   = read_->first;   // Read the key from the iterator
  val_name   = read_->second;  // Read the value from the iterator
  iterating_ = true;

  // Verify buffer
  if (*name_len < val_name.length() + 1) {
    // The buffer is not large enough
    *name_len = val_name.length() + 1;
	// Unlock list
	LeaveCriticalSection(&lock_);
	return 2;
  }

  // move iterator to next node
  read_++;

  // return node and name
  *node     = val_node;
  strcpy_s(name, *name_len, val_name.c_str());
  *name_len = val_name.length() + 1;

  // Unlock list
  LeaveCriticalSection(&lock_);

  // Completed
  return 0;
}

/////////////////////////////////////////////////////////////////////////////
// This function is called from the RCH when iterate the names in the
// list, after the iteration is started.
// The list will be locked until iteration is completed.
//
// Input:
//    name_len      - Pointer to a variable that holds the length of the
//                    name buffer in Bytes. If the buffer is to small it
//                    will hold the size required.
//
// Output:
//    name          - Pointer to the buffer where the name is stored.
//                    The name is a zero terminated UTF-8 string.
//    node          - Pointer to the buffer where the Node ID is stored.
//
// Returns:
//    0             - The name and node id is returned.
//    1             - One or more of the parameters are missing.
//    2             - The name buffer is not large enough.
//    3             - No more names in list.
//
/////////////////////////////////////////////////////////////////////////////
int  GetNextName(char *name, int *name_len, long *node)
{
  // Var
  long          val_node;
  std::string   val_name;

  // Verify inputs
  if (name == nullptr) return 1;
  if (name_len == nullptr) return 1;
  if (node == nullptr) return 1;

  // Verify read iterator
  if (!iterating_) return 3;
  if (read_ == list_.end()) {
    // Unlock list
    LeaveCriticalSection(&lock_);
    iterating_ = false;
    // No names in list
    return 3;
  }

  // Get next entry
  val_node  = read_->first;   // Read the key from the iterator
  val_name  = read_->second;  // Read the value from the iterator

  // Verify buffer
  if (*name_len < val_name.length() + 1) {
    // The buffer is not large enough
    *name_len = val_name.length() + 1;
    return 2;
  }

  // move iterator to next node
  read_++;

  // return node and name
  *node     = val_node;
  strcpy_s(name, *name_len, val_name.c_str());
  *name_len = val_name.length() + 1;
  return 0;
}


/****************************************************************************
**
**    EOF 
**
****************************************************************************/
