
#ifndef _GATEWAY_SERVICES_H_
#define _GATEWAY_SERVICES_H_

#ifdef __cplusplus
  extern "C" {
#endif


//
// Log service
//
#define GWPF_SERVICE_LOG                2

/////////////////////////////////////////////////////////////////////////////
// This structure holds the parameters for inserting a message into the
// Gateway log.
//
// Input:
//    type              - What type of message is it.
//                        (1 = Error, 2 = Event, 3 = Detailed Event)
//
//    message           - Pointer to zero terminated UTF-8 text.
//
// Output:
//    None.
//
// Returns:
//    No additional errors.
//
/////////////////////////////////////////////////////////////////////////////
typedef struct gwpf_service_log_params {
   int                type;
   const char         *message;
} gwpf_service_log_params;


//
// Logon service
//
#define GWPF_SERVICE_LOGON              3

/////////////////////////////////////////////////////////////////////////////
// This is the function pointer used by the Gateway to inform the plugin
// about a client login attempt. The plugin is informed before the gateway
// checks key and max client limit.
//
// Input:
//    NodeID            - The requested node id. (zero = dynamic)
//
//    addr              - A zero terminated string with the IP address of
//                        client. The address is in numerical form.
//                        (For example: 192.168.0.1)
//
//    port              - The IP port of the client.
//
// Output:
//    None.
//
// Returns:
//    0                 - Client logon is denied by plugin.
//    1                 - Client logon is allowed by plugin.
//
/////////////////////////////////////////////////////////////////////////////
typedef int (__cdecl *gwpf_func_onlogon)(const long           NodeID,
                                         const char           *addr,
                                         const unsigned short port);

/////////////////////////////////////////////////////////////////////////////
// This structure holds the parameters for subscribing to the client
// logon attempt.
//
// Input:
//    func              - Pointer to function that is called to notify plugin.
//
// Output:
//    None.
//
// Returns:
//    No additional errors.
//
/////////////////////////////////////////////////////////////////////////////
typedef struct gwpf_service_onlogon_params {
   gwpf_func_onlogon   func;
} gwpf_service_onlogon_params;


//
// Connected service
//
#define GWPF_SERVICE_CONNECT            4

/////////////////////////////////////////////////////////////////////////////
// This is the function pointer used by the Gateway to inform the plugin
// about a successfull logon by a client.
//
// Input:
//    NodeID            - The node id of the client that is connected.
//
// Output:
//    None.
//
// Returns:
//    None.
//
/////////////////////////////////////////////////////////////////////////////
typedef void (__cdecl *gwpf_func_onconnect)(const long NodeID);

/////////////////////////////////////////////////////////////////////////////
// This structure holds the parameters for subscribing to the client
// connected event.
//
// Input:
//    func              - Pointer to function that is called to notify plugin.
//
// Output:
//    None.
//
// Returns:
//    No additional errors.
//
/////////////////////////////////////////////////////////////////////////////
typedef struct gwpf_service_onconnect_params {
   gwpf_func_onconnect func;
} gwpf_service_onconnect_params;


//
// Closed service
//
#define GWPF_SERVICE_CLOSED             5

/////////////////////////////////////////////////////////////////////////////
// This is the function pointer used by the Gateway to inform the plugin
// that a client is closed / disconnected.
//
// Input:
//    NodeID            - The node id of the client that is closed.
//
// Output:
//    None.
//
// Returns:
//    None.
//
/////////////////////////////////////////////////////////////////////////////
typedef void (__cdecl *gwpf_func_onclose)(const long NodeID);

/////////////////////////////////////////////////////////////////////////////
// This structure holds the parameters for subscribing to the client
// closed event.
//
// Input:
//    func              - Pointer to function that is called to notify plugin.
//
// Output:
//    None.
//
// Returns:
//    No additional errors.
//
/////////////////////////////////////////////////////////////////////////////
typedef struct gwpf_service_onclose_params {
   gwpf_func_onclose func;
} gwpf_service_onclose_params;


//
// Name list service
//
#define GWPF_SERVICE_NAMELIST           6

/////////////////////////////////////////////////////////////////////////////
// This is the function pointer used by the Gateway to request a name and
// node id pair from the plugin when iterating the name list.
//
// Input:
//    name_len          - Pointer to a variable that holds the length of the
//                        name buffer in Bytes. If the buffer is to small it
//                        will hold the size required.
//
// Output:
//    name              - Pointer to the buffer where the name is stored.
//                        The name is a zero terminated UTF-8 string.
//    node              - Pointer to the buffer where the Node ID is stored.
//
// Returns:
//    0                 - The name and node id is returned.
//    1                 - One or more of the parameters are missing.
//    2                 - The name buffer is not large enough.
//    3                 - No more names in list.
//
/////////////////////////////////////////////////////////////////////////////
typedef int (__cdecl *gwpf_func_nameitem)(char *name, int *name_len, long *node);

/////////////////////////////////////////////////////////////////////////////
// This structure holds the parameters for subscribing to the name list event.
//
// Input:
//    first             - Pointer to function that is called to get the first
//                        entry in the name list.
//    next              - Pointer to function that is called to get entries
//                        from the name list after the first one.
//    lookup            - Pointer to function that is called to lookup the
//                        name of a particular Node ID.
//
// Output:
//    None.
//
// Returns:
//    No additional errors.
//
/////////////////////////////////////////////////////////////////////////////
typedef struct gwpf_service_namelist_params {
   gwpf_func_nameitem       first;
   gwpf_func_nameitem       next;
} gwpf_service_namelist_params;


#ifdef  __cplusplus
   }
#endif

#endif  /* _GATEWAY_SERVICES_H_ */
